#pragma once

#include <osmocom/core/logging.h>

/* Debug Areas of the code */
enum {
	DRLL,
	DCC,
	DBCC,
	DGCC,
	DMM,
	DRR,
	DLCLS,
	DMNCC,
	DPAG,
	DMSC,
	DMGCP,
	DHO,
	DDB,
	DREF,
	DCTRL,
	DSMPP,
	DRANAP,
	DVLR,
	DIUCS,
	DBSSAP,
	DSGS,
	DSS,
	DASCI,
	Debug_LastEntry,
};
