#ifndef _GSM_SUBSCR_H
#define _GSM_SUBSCR_H

#include <stdbool.h>

#include <osmocom/core/linuxlist.h>
#include <osmocom/gsm/protocol/gsm_23_003.h>
#include <osmocom/gsm/protocol/gsm_29_118.h>

#include <osmocom/msc/gsm_data.h>

struct ran_conn;
struct msgb;

#endif /* _GSM_SUBSCR_H */
